
from .LieGroup import LieGroup
import numpy as np
from .R3 import R3 as R3

class MR1(LieGroup):
    # The Lie group of dim 1 scaling.
    # [ s ]
    def __init__(self, s = 1.0):
        self._scale = s
    
    def scale(self):
        return self._scale
    
    def Adjoint(self):
        return np.eye(1)
    
    @staticmethod
    def adjoint(mr1vec):
        return np.array([[0]])
    
    def __mul__(self, other):
        if isinstance(other, MR1):
            result = MR1()
            result._scale = self._scale * other._scale
            return result
        elif isinstance(other, np.ndarray):
            return float(self._scale) * other
        elif isinstance(other, float):
            return float(self._scale) * other
        elif isinstance(other, R3):
            result = R3()
            result._trans = self._scale * other._trans
            return result
        return NotImplemented
    
    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __truediv__(self, other):
        if isinstance(other, MR1):
            result = MR1()
            result._scale = self._scale / other._scale
            return result
        elif isinstance(other, float):
            result = self / MR1(other)
            return result
        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, MR1):
            result = MR1()
            result._scale = other._scale / self._scale
            return result
        elif isinstance(other, np.ndarray):
            return other / self.as_float()
        elif isinstance(other, float):
            result = MR1(other) / self
            return result
        return NotImplemented
    
    def inv(self):
        result = MR1()
        result._scale = 1.0/self._scale
        return result
    
    def log(self):
        return np.log(self._scale)
    
    def as_matrix(self):
        return np.array(self._scale)
    
    def as_float(self):
        return float(self._scale)
    
    def __float__(self):
        return float(self._scale)
    
    @staticmethod
    def identity():
        result = MR1()
        result._scale = 1.0
        return result

    @staticmethod
    def exp(mr1vec):
        if isinstance(mr1vec, np.ndarray):
            assert mr1vec.size == 1, "The Lie algebra element should hold only 1 element."
            mr1vec = float(mr1vec)
        result = MR1()
        result._scale = np.exp(mr1vec)
        return result

    @staticmethod
    def valid_list_formats() -> dict:
        # Possible formats are
        # s : 1 entry scale
        return {'s':1}

    @staticmethod
    def from_list(line, format_spec="s") -> 'MR1':
        result = MR1()
        if format_spec == "s":
            result._scale = float(line[0])
        else:
            return NotImplemented
        return result

    def to_list(self, format_spec) -> list:
        if format_spec == "s":
            result = [float(self._scale)]
        else:
            return NotImplemented
        return result
    
    @staticmethod
    def list_header(format_spec):
        if format_spec == "s":
            result = ["s"]
        else:
            return NotImplemented
        return result

if __name__ == "__main__":
    s = MR1()



from .LieGroup import LieGroup
import numpy as np

class SE2(LieGroup):
    DIM = 3
    _cross_matrix = np.array(((0.,-1.),(1.,0.)))
    def __init__(self, theta : float = 0., p : np.ndarray = None):
        if p is None:
            p = np.zeros(2)
        self._theta = SE2._wrap_angle(theta)
        self._p = p
    
    def theta(self) -> float:
        return self._theta
    
    def R(self) -> np.ndarray:
        return SE2._rotation_matrix(self._theta)
    
    @staticmethod
    def _rotation_matrix(theta):
        c = np.cos(theta)
        s = np.sin(theta)
        return np.array((
            (c, -s),
            (s, c)
        ))
    
    @staticmethod
    def _wrap_angle(t):
        return t - np.floor((t + np.pi) / (2*np.pi)) * (2*np.pi)

    
    def p(self) -> np.ndarray:
        return self._p

    def __str__(self) -> str:
        return str(self.as_matrix())
    
    def Adjoint(self) -> np.ndarray:
        Ad = np.eye(3)
        Ad[1:,0] = -self._cross_matrix @ self._p
        Ad[1:,1:] = self.R()
        return Ad
    
    @staticmethod
    def adjoint(se2vec : np.ndarray) -> np.ndarray:
        assert isinstance(se2vec, np.ndarray)
        assert se2vec.size == 3
        ad = np.zeros((3,3),dtype=se2vec.dtype)
        ad[1:,0] = -SE2._cross_matrix @ se2vec[1:]
        ad[1:,1:] = SE2._cross_matrix * se2vec[0]
        return ad
    
    def __mul__(self, other) -> 'SE2':
        if isinstance(other, SE2):
            result = SE2()
            result._theta = SE2._wrap_angle(self._theta + other._theta)
            result._p = self._p + self.R() @ other._p
            return result
        if isinstance(other, np.ndarray):
            if other.shape[0] == 2:
                return self._p + self.R() @ other
            elif other.shape[0] == 3:
                return self.as_matrix() @ other
        
        return NotImplemented
    
    @staticmethod
    def identity() -> 'SE2':
        result = SE2()
        result._theta = 0.
        result._p = np.zeros(2)
        return result
    
    def as_matrix(self) -> np.ndarray:
        mat = np.eye(3)
        mat[0:2,0:2] = self.R()
        mat[0:2,2] = self._p
        return mat

    @staticmethod
    def from_matrix(mat : np.ndarray) -> 'SE2':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (3,3):
            raise ValueError
        
        result = SE2()
        result._theta = np.arctan2(mat[1,0], mat[0,0])
        result._p = mat[0:2,2]
        return result
    
    def as_vector(self) -> np.ndarray:
        return np.array([self._theta, self._p[0], self._p[1]])
    
    @staticmethod
    def from_vector(mat : np.ndarray) -> 'SE2':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.size == 3:
            raise ValueError
        
        result = SE2()
        result._theta = mat.item(0)
        result._p = np.array([mat.item(1),mat.item(2)])
        return result

    def __truediv__(self, other) -> 'SE2':
        if isinstance(other, SE2):
            return self * other.inv()
        return NotImplemented
    
    def inv(self) -> 'SE2':
        result = SE2()
        result._theta = - self._theta
        result._p = - self.R().T @ self._p
        return result
    
    @staticmethod
    def exp(se2arr) -> 'SE2':
        if not isinstance(se2arr, np.ndarray):
            raise TypeError
        if se2arr.shape == (3,3):
            se2arr = SE2.vee(se2arr)
        elif not se2arr.size == 3:
            raise ValueError

        w = se2arr[0]

        if abs(w) > 1e-6:
            A = np.sin(w) / w
            B = (1.0 - np.cos(w)) / w**2.0
        else:
            A = 1.0
            B = 1.0 / 2.0
        
        U = SE2.wedge(se2arr)
        

        mat = np.eye(3) + A * U + B * U @ U
        result = SE2.from_matrix(mat)
        result._theta = w
        return result
    
    def log(self) -> np.ndarray:
        w = self._theta
        if abs(w) > 1e-6:
            v = w / (2*(1-np.cos(w))) * self._cross_matrix.T @ (np.eye(2) - self.R().T) @ self._p
        else:
            v = self._p
        return np.concatenate(([w], v))
        
    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (3,3):
            raise ValueError
        vecOmega = -mat[0,1]
        vecV = mat[0:2,2]
        return np.concatenate(([vecOmega], vecV))

    @staticmethod
    def wedge(vec : np.ndarray) -> np.ndarray:
        if not isinstance(vec, np.ndarray):
            raise TypeError
        if not vec.size == 3:
            raise ValueError
        mat = np.zeros((3,3),dtype=vec.dtype)
        mat[0:2,0:2] = vec[0] * SE2._cross_matrix
        mat[0:2,2] = vec.ravel()[1:]
        return mat

if __name__ == "__main__":
    P = SE2()


import pylie
import numpy as np


class Trajectory:
    def __init__(self, elements=None, times=None):
        if elements is None:
            self._elements = []
            self._times = []
        else:
            self._elements = elements

        if times is None:
            self._times = list(range(len(self._elements)))
        else:
            self._times = times

        self._clean_duplicates()

    def __len__(self):
        assert len(self._elements) == len(self._times)
        return len(self._elements)

    def __getitem__(self, t):
        if isinstance(t, (int, float, complex)) and not isinstance(t, bool):
            # t is a number
            t = float(t)
            i0, i1 = self._get_indices_near_time(t)
            if i0 is None:
                return None
            elif i1 is None:
                return self._elements[0]

            # Now (inter/extra)polate
            X0 = self._elements[i0]
            t0 = self._times[i0]
            X1 = self._elements[i1]
            t1 = self._times[i1]

            dt = t1 - t0
            motion = (X0.inv() * X1).log() / dt
            elem = X0 * X0.exp((t - t0) * motion)
            return elem

        elif isinstance(t, list):
            # t is a list of numbers -> return the corresponding trajectory
            return Trajectory([self[tau] for tau in t], t)

        raise NotImplementedError

    def get_velocity(self, t):
        if isinstance(t, (int, float, complex)) and not isinstance(t, bool):
            # t is a number
            t = float(t)
            i0, i1 = self._get_indices_near_time(t)
            if i0 is None:
                return None
            elif i1 is None:
                return self._elements[0].log() * 0.0

            # Now (inter/extra)polate
            X0 = self._elements[i0]
            t0 = self._times[i0]
            X1 = self._elements[i1]
            t1 = self._times[i1]

            motion = (X0.inv() * X1).log() / (t1 - t0)

            assert not np.isnan(motion).any()
            return motion

        raise NotImplementedError

    def get_velocities(self):
        vels = [
            (self._elements[i].inv() * self._elements[i+1]).log() / (self._times[i+1] - self._times[i]) for i in range(len(self)-1)
        ]
        return vels

    def begin_time(self):
        return self._times[0]

    def end_time(self):
        return self._times[-1]

    def get_elements(self):
        return self._elements

    def get_times(self):
        return self._times

    def group_type(self):
        if len(self._elements) > 0:
            return type(self._elements[0])
        else:
            return None

    def __rmul__(self, other):
        if isinstance(other, self.group_type()):
            new_elements = [other * X for X in self._elements]
            new_times = [t for t in self._times]
            return Trajectory(new_elements, new_times)
        raise NotImplementedError

    def __mul__(self, other):
        if isinstance(other, self.group_type()):
            new_elements = [X * other for X in self._elements]
            new_times = [t for t in self._times]
            return Trajectory(new_elements, new_times)
        raise NotImplementedError

    def truncate(self, t0, t1):
        assert t0 < t1

        idx0 = [j for j in range(len(self._times)) if self._times[j] >= t0]
        if len(idx0) > 0:
            idx0 = idx0[0]
        else:
            self._times.clear()
            self._elements.clear()
            return

        idx1 = [j for j in reversed(
            range(len(self._times))) if self._times[j] <= t1]
        if len(idx1) > 0:
            idx1 = idx1[0]
        else:
            self._times.clear()
            self._elements.clear()
            return

        self._times = self._times[idx0:idx1]
        self._elements = self._elements[idx0:idx1]

    def _clean_duplicates(self):
        for i in reversed(range(len(self._times)-1)):
            if self._times[i+1] - self._times[i] < 1e-8:
                del self._times[i+1]
                del self._elements[i+1]

    def _get_indices_near_time(self, t: float):
        assert(isinstance(t, float))
        if len(self._times) == 0:
            return None, None
        elif len(self._times) == 1:
            return 0, None

        next_idx = [j for j in range(len(self._times)) if self._times[j] > t]
        if len(next_idx) == 0:
            # There are no times after t
            return len(self._times)-2, len(self._times)-1
        elif next_idx[0] == 0:
            # All times are after t
            return 0, 1
        else:
            return next_idx[0]-1, next_idx[0]


if __name__ == "__main__":
    import numpy as np
    times = [i * 0.1 for i in range(10)]
    motion = np.random.randn(3, 1)
    elements = [pylie.SO3.exp(motion*t) for t in times]
    traj = Trajectory(elements, times)

    new_times = [0.2, 0.45, 1.2, -0.12]
    for t in new_times:
        group_error = traj[t] / pylie.SO3.exp(t*motion)
        norm_error = np.linalg.norm(group_error.as_matrix() - np.eye(3))
        print("The error at time {} is {}".format(t, norm_error))



from .LieGroup import LieGroup
import numpy as np

# The indefinite orthogonal group of signature 1,2
class O1_2(LieGroup):
    DIM = 3
    _cross_mat = np.array(((0,-1),(1,0)))
    def __init__(self, L : np.ndarray = None):
        if L is None:
            self._L = np.eye(3)
        else:
            assert L.shape == (3,3), "L must be a 3x3 matrix"
            self._L = L
    
    def L(self) -> np.ndarray:
        return self._L
    
    def _extract_components(self):
        return self._L[1:,1:], self._L[0,1:], self._L[1:,0], self._L[0,0]
    
    def _validity(self):
        # Check how valid a given matrix is. Returns a nonnegative float that should be zero.
        g = np.eye(3)
        g[1:,1:] = -np.eye(2)
        error_mat = self._L.T @ g @ self._L - g
        return np.linalg.norm(error_mat)


    def __str__(self) -> str:
        return str(self.as_matrix())
    
    def Adjoint(self) -> np.ndarray:
        Ad_mat = np.empty((3,3))
        A,b,c,d = self._extract_components()
        Ad_mat[0,0] = -0.5 * np.trace(self._cross_mat @ A @ self._cross_mat @ A.T)
        Ad_mat[1:,0] = -A @ self._cross_mat @ b
        Ad_mat[0,1:] = A.T @ self._cross_mat.T @ c
        Ad_mat[1:,1:] = d * A - np.outer(c,b)
        return Ad_mat
        
    
    @staticmethod
    def adjoint(o1_2vec : np.ndarray) -> np.ndarray:
        assert isinstance(o1_2vec, np.ndarray)
        assert o1_2vec.size == 3
        ad_mat = np.empty((3,3))
        ad_mat[0,0] = 0.0
        omega = o1_2vec[0]; u = o1_2vec[1:]
        ad_mat[1:,0] = - O1_2._cross_mat @ u
        ad_mat[0,1:] = O1_2._cross_mat.T @ u
        ad_mat[1:,1:] = O1_2._cross_mat * omega
        return ad_mat
    
    def __mul__(self, other) -> 'O1_2':
        if isinstance(other, O1_2):
            result = O1_2(self._L @ other._L)
            return result
        if isinstance(other, np.ndarray):
            if other.shape[0] == 3:
                return self.L @ other
        
        return NotImplemented
    
    @staticmethod
    def identity() -> 'O1_2':
        return O1_2(np.eye(3))
    
    def as_matrix(self) -> np.ndarray:
        return self._L

    @staticmethod
    def from_matrix(mat : np.ndarray) -> 'O1_2':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (3,3):
            raise ValueError
        
        result = O1_2(mat)
        return result
    
    def __truediv__(self, other) -> 'O1_2':
        if isinstance(other, O1_2):
            return self * other.inv()
        return NotImplemented
    
    def inv(self) -> 'O1_2':
        A,b,c,d = self._extract_components()
        LInv = np.empty((3,3))
        LInv[1:,1:] = A.T
        LInv[0,0] = d
        LInv[0,1:] = -c
        LInv[1:,0] = -b
        return O1_2(LInv)
    
    @staticmethod
    def exp(O1_2vec) -> 'O1_2':
        if not isinstance(O1_2vec, np.ndarray):
            raise TypeError
        if O1_2vec.shape == (3,3):
            u = O1_2.vee(O1_2vec)
        elif O1_2vec.size == 3:
            u = O1_2vec
        else:
            raise ValueError
        
        q2 = u[1]**2 + u[2]**2 - u[0]**2

        U = O1_2.wedge(u)
        if q2 == 0.:
            L = np.eye(3) + U + 0.5 * U@U
        elif q2 > 0:
            q = np.sqrt(q2)
            L = np.eye(3) + np.sinh(q) / q * U + (np.cosh(q) - 1)/q2 * U@U
        else:
            iq = np.sqrt(-q2)
            L = np.eye(3) + np.sin(iq) / iq * U + (np.cos(iq) - 1)/q2 * U@U

        return O1_2(L)
    
    def log(self) -> np.ndarray:
        A,b,c,d = self._extract_components()
        sinh_q_2 = np.linalg.norm(0.5*(b+c))**2 - 0.25 * (A[0,1] - A[1,0])**2
        if sinh_q_2 == 0:
            term_1 = 0
        elif sinh_q_2 > 0:
            q = np.arcsinh(np.sqrt(sinh_q_2))
            term_1 = np.sinh(q) / q
        else:
            iq = np.arcsin(np.sqrt(-sinh_q_2))
            term_1 = np.sin(iq)/iq
        
        U = np.empty(3)
        U[0] = (A[1,0]-A[0,1]) / (2*term_1)
        U[1:] = (b+c) / (2*term_1)

        return U
        
    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (3,3):
            raise ValueError

        return np.array((-mat[1,2], mat[0,1], mat[0,2]))

    @staticmethod
    def wedge(vec : np.ndarray) -> np.ndarray:
        if not isinstance(vec, np.ndarray):
            raise TypeError
        if not vec.size == 3:
            raise ValueError
        mat = np.array((
            (0, vec.item(1), vec.item(2)),
            (vec.item(1), 0, -vec.item(0)),
            (vec.item(2), vec.item(0), 0)
        ))
        return mat

if __name__ == "__main__":
    P = O1_2()


from .LieGroup import LieGroup
from .SO3 import SO3
from .SOT3 import SOT3
from .SE3 import SE3
from .SE23 import SE23
from .SIM3 import SIM3
from .MR1 import MR1
from .R3 import R3
from .SE2 import SE2
from .SL2 import SL2
from .O1_2 import O1_2
from .Quaternion import Quaternion
from .Trajectory import Trajectory
from . import analysis
from . import plotting


from .SIM3 import SIM3
import numpy as np
from .Trajectory import Trajectory

def align_trajectory(trajectory0 : Trajectory, trajectory1 : Trajectory, ret_params=False) -> Trajectory:
    # Align trajectory0 to trajectory1
    assert isinstance(trajectory0, Trajectory)
    assert isinstance(trajectory1, Trajectory)
    
    t0 = max(trajectory0.begin_time(), trajectory1.begin_time())
    t1 = min(trajectory0.end_time(), trajectory1.end_time())
    if t0 >= t1:
        if ret_params:
            return Trajectory(), SIM3.identity()
        else:
            return Trajectory()
    
    num_points = 100
    times = np.linspace(t0, t1, num_points).tolist()
    points0 = [trajectory0[t].x().as_vector() for t in times]
    points1 = [trajectory1[t].x().as_vector() for t in times]

    S = umeyama(points0, points1)

    trajectoryA = S.to_SE3() * trajectory0
    if ret_params:
        return trajectoryA, S
    else:
        return trajectoryA



def umeyama(points1 : np.ndarray, points2 : np.ndarray) -> SIM3:
    # This function solves the least squares problem of finding a SIM3 transform S such that
    # S * points1 = points2,
    # s_S * R_S * points1_i + x_S = points2_i
    def _list_to_stack(points):
        if isinstance(points, list):
            return np.stack([p.ravel() for p in points]).T
        else:
            return points
    points1 = _list_to_stack(points1)
    points2 = _list_to_stack(points2)
    assert isinstance(points1, np.ndarray) and isinstance(points2, np.ndarray), "The points must be in numpy arrays."
    assert points1.shape == points2.shape, "The points are not matched."
    assert points1.shape[0] == 3, "The points are not 3D."

    # Compute relevant variables
    n = points1.shape[1]
    mu1 = 1/n * np.reshape(np.sum(points1, axis=1),(3,-1))
    mu2 = 1/n * np.reshape(np.sum(points2, axis=1),(3,-1))
    sig1sq = 1/n * np.sum((points1 - mu1)**2.0)
    sig2sq = 1/n * np.sum((points2 - mu2)**2.0)
    Sig12 = 1/n * (points2-mu2) @ (points1-mu1).T

    # Use the SVD for the rotation
    U, d, Vh = np.linalg.svd(Sig12)
    S = np.eye(3)
    if np.linalg.det(Sig12) < 0:
        S[-1,-1] = -1
    
    R = U @ S @ Vh
    s = 1.0 / sig1sq * np.trace(np.diag(d) @ S)
    x = mu2 - s * R @ mu1

    # Return the result as a SIM3 element
    result = SIM3()
    result._R = result._R.from_matrix(R)
    result._x.__init__(x)
    result._s.__init__(float(s))

    return result

if __name__ == "__main__":
    true_transform = SIM3.from_list(np.random.randn(1,7).ravel().tolist(), "srx")
    print(true_transform)

    points1 = np.random.randn(3,10)
    points2 = true_transform * points1

    est_transform = umeyama(points1, points2)
    print(est_transform)
    


from .LieGroup import LieGroup
from .SE3 import SE3 as SE3
from .SO3 import SO3 as SO3
from .R3 import R3 as R3
from .MR1 import MR1 as MR1
import numpy as np


class SIM3(LieGroup):
    DIM = 7
    # SIM(3) is defined with the matrix form
    # [ sR x ]
    # [ 0  1 ]
    def __init__(self, R=SO3(), x=R3(), s=MR1()):
        self._R = R
        self._x = x
        self._s = s

    def R(self):
        return self._R

    def x(self):
        return self._x

    def s(self):
        return self._s

    def __str__(self):
        return str(self.as_matrix())

    def Adjoint(self):
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, SIM3):
            result = SIM3()
            result._R = self._R * other._R
            result._x = self._x + self._s * (self._R * other._x)
            result._s = self._s * other._s
            return result
        elif isinstance(other, np.ndarray):
            if other.shape[0] == 3:
                return self._x + self._s * (self._R * other)
            elif other.shape[0] == 4:
                return self.as_matrix() @ other
        elif isinstance(other, SE3):
            result = SIM3()
            result._R = self._R * other._R
            result._x = self._x + self._s * (self._R * other._x)
            result._s = self._s
            return result

        return NotImplemented

    def as_matrix(self):
        mat = np.eye(4)
        mat[0:3, 0:3] = self._s._scale * self._R.as_matrix()
        mat[0:3, 3] = self._x._trans
        return mat
    
    def to_SE3(self):
        result = SE3()
        result._R = self._R
        result._x = self._x
        return result

    def __truediv__(self, other):
        if isinstance(other, SIM3):
            return self * other.inv()
        return NotImplemented

    def inv(self):
        result = SIM3()
        result._R = self._R.inv()
        result._x = - (self._s.inv() * (self._R.inv() * self._x))
        result._s = self._s.inv()
        return result

    def log(self):
        return NotImplemented

    @staticmethod
    def exp(sim3vec):
        assert sim3vec.shape == (7, 1), "Invalid shape of Lie algebra vector."
        return NotImplemented

    @staticmethod
    def identity() -> 'SIM3':
        result = SIM3()
        result._R = SO3.identity()
        result._x = R3.identity()
        result._s = MR1.identity()
        return result

    @staticmethod
    def valid_list_formats():
        # Possible formats are
        # SO(3) format specs
        # R(3) format specs
        # S(1) format specs
        result = dict()
        result.update(SO3.valid_list_formats())
        result.update(R3.valid_list_formats())
        result.update(MR1.valid_list_formats())
        return result

    @staticmethod
    def from_list(line, format_spec="sqx") -> 'SE3':
        result = SE3()
        SO3_formats = SO3.valid_list_formats()
        R3_formats = R3.valid_list_formats()
        MR1_formats = MR1.valid_list_formats()
        for fspec in format_spec:
            if fspec in SO3_formats:
                result._R = SO3.from_list(line, fspec)
            elif fspec in R3_formats:
                result._x = R3.from_list(line, fspec)
            elif fspec in MR1_formats:
                result._s = MR1.from_list(line, fspec)
            else:
                return NotImplemented
        return result

    def to_list(self, format_spec) -> list:
        result = []
        SO3_formats = SO3.valid_list_formats()
        R3_formats = R3.valid_list_formats()
        MR1_formats = MR1.valid_list_formats()
        for fspec in format_spec:
            if fspec in SO3_formats:
                result += self._R.to_list(fspec)
            elif fspec in R3_formats:
                result += self._x.to_list(fspec)
            elif fspec in MR1_formats:
                result += self._s.to_list(fspec)
            else:
                return NotImplemented
        return result

    @staticmethod
    def list_header(format_spec) -> list:
        result = []
        SO3_formats = SO3.valid_list_formats()
        R3_formats = R3.valid_list_formats()
        MR1_formats = MR1.valid_list_formats()
        for fspec in format_spec:
            if fspec in R3_formats:
                result += R3.list_header(fspec)
            elif fspec in SO3_formats:
                result += SO3.list_header(fspec)
            elif fspec in MR1_formats:
                result += MR1.list_header(fspec)
            else:
                return NotImplemented
        return result


if __name__ == "__main__":
    S = SIM3()



from .LieGroup import LieGroup
from .SO3 import SO3 as SO3
from .SE3 import SE3 as SE3
from .R3 import R3 as R3
import numpy as np

class SE23(LieGroup):
    DIM = 9
    def __init__(self, R : SO3 = None, x : R3 = None, w : R3 = None):
        if R is None:
            R = SO3()
        if x is None:
            x = R3()
        if w is None:
            w = R3()
        self._R = SO3(R)
        self._x = R3(x)
        self._w = R3(w)
    
    def R(self) -> SO3:
        return self._R
    
    def x(self) -> R3:
        return self._x
    
    def P(self) -> SE3:
        return SE3(self._R, self._x)

    def w(self) -> R3:
        return self._w

    def __str__(self) -> str:
        return str(self.as_matrix())
    
    def Adjoint(self) -> np.ndarray:
        Ad = np.zeros((9,9))
        R = self.R().as_matrix()
        Ad[0:3,0:3] = R
        Ad[3:6,0:3] = SO3.skew(self.x().as_vector()) @ R
        Ad[3:6,3:6] = R
        Ad[6:9,0:3] = SO3.skew(self.w().as_vector()) @ R
        Ad[6:9,6:9] = R
        return Ad
    
    @staticmethod
    def adjoint(se23vec : np.ndarray) -> np.ndarray:
        assert isinstance(se23vec, np.ndarray)
        assert se23vec.size == 9
        ad = np.zeros((9,9))
        OmegaCross = SO3.skew(se23vec.ravel()[0:3])
        VCross = SO3.skew(se23vec.ravel()[3:6])
        ACross = SO3.skew(se23vec.ravel()[6:9])
        ad[0:3,0:3] = OmegaCross
        ad[3:6,0:3] = VCross
        ad[3:6,3:6] = OmegaCross
        ad[6:9,0:3] = ACross
        ad[6:9,6:9] = OmegaCross
        return ad
    
    def __mul__(self, other) -> 'SE23':
        if isinstance(other, SE23):
            result = SE23()
            result._R = self._R * other._R
            result._x = self._x + (self._R * other._x)
            result._w = self._w + (self._R * other._w)
            return result
        if isinstance(other, np.ndarray):
            if other.shape[0] == 5:
                return self.as_matrix() @ other
        
        return NotImplemented
    
    @staticmethod
    def identity() -> 'SE23':
        result = SE23()
        result._R = SO3.identity()
        result._x = R3.identity()
        result._w = R3.identity()
        return result
    
    def as_matrix(self) -> np.ndarray:
        mat = np.eye(5)
        mat[0:3,0:3] = self._R.as_matrix()
        mat[0:3,3] = self._x.as_vector()
        mat[0:3,4] = self._w.as_vector()
        return mat

    @staticmethod
    def from_matrix(mat : np.ndarray) -> 'SE23':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (5,5):
            raise ValueError
        
        result = SE23()
        result._R = SO3.from_matrix(mat[0:3,0:3])
        result._x = R3(mat[0:3,3:4])
        result._w = R3(mat[0:3,4:5])
        return result
    
    def __truediv__(self, other) -> 'SE23':
        if isinstance(other, SE23):
            return self * other.inv()
        return NotImplemented
    
    def inv(self) -> 'SE23':
        result = SE23()
        result._R = self._R.inv()
        result._x = - (self._R.inv() * self._x)
        result._w = - (self._R.inv() * self._w)
        return result
    
    @staticmethod
    def exp(se3arr) -> 'SE23':
        if not isinstance(se3arr, np.ndarray):
            raise TypeError
        if se3arr.shape == (5,5):
            se3arr = SE23.vee(se3arr)
        elif not se3arr.size == 9:
            raise ValueError

        w = se3arr.ravel()[0:3]
        theta = np.linalg.norm(w)

        if theta > 1e-6:
            A = np.sin(theta) / theta
            B = (1.0 - np.cos(theta)) / theta**2.0
            C = (1.0 - A) / theta**2.0
        else:
            A = 1.0
            B = 1.0 / 2.0
            C = 1.0 / 6.0
        
        wx = SO3.skew(w)
        wx2 = wx @ wx
        R = np.eye(3) + A * wx + B * wx2
        V = np.eye(3) + B * wx + C * wx2

        u1 = se3arr.ravel()[3:6]
        u2 = se3arr.ravel()[6:9]
        result = SE23(R, V@u1, V@u2)

        return result
    
    def log(self) -> np.ndarray:
        w = self._R.log()
        theta = np.linalg.norm(w)
        wx = SO3.skew(w)
        if theta > 1e-6:
            Vinv = np.eye(3) - 0.5 * wx + theta**(-2.0) * (1.0 - (theta*np.sin(theta))/(2*(1-np.cos(theta)))) * wx @ wx
        else:
            Vinv = np.eye(3) - 0.5*wx
        u1 = Vinv @ self._x.as_vector()
        u2 = Vinv @ self._w.as_vector()
        return np.concatenate((w,u1,u2))

    # @staticmethod
    # def valid_list_formats() -> dict:
    #     # Possible formats are
    #     # q/w/R/r : SO(3) format specs
    #     # x : 3 entry translation
    #     # v : 3 entry velocity
    #     # P : 12 entry homogeneous matrix (row-by-row)
    #     result = {'P':12, 'v':3}
    #     result.update(SO3.valid_list_formats())
    #     result.update(R3.valid_list_formats())
    #     return result

    # @staticmethod
    # def from_list(line, format_spec="qx") -> 'SE3':
    #     result = SE3()
    #     SO3_formats = SO3.valid_list_formats()
    #     R3_formats = R3.valid_list_formats()
    #     for fspec in format_spec:
    #         if fspec in SO3_formats:
    #             result._R = SO3.from_list(line, fspec)
    #             line = line[SO3_formats[fspec]:]
    #         elif fspec in R3_formats:
    #             result._x = R3.from_list(line, fspec)
    #             line = line[R3_formats[fspec]:]
    #         elif fspec == "P":
    #             mat = np.reshape(np.array([float(line[i]) for i in range(12)]), (3,4))
    #             result._R = result._R.from_matrix(mat[0:3,0:3])
    #             result._x = R3(mat[0:3,3:4])
    #             line = line[12:]
    #         elif fspec == "v":
    #             result._w = R3.from_list(line, 'x')
    #             line = line[R3_formats['x']:]
    #         else:
    #             return NotImplemented
    #     return result

    # def to_list(self, format_spec) -> list:
    #     result = []
    #     SO3_formats = SO3.valid_list_formats()
    #     R3_formats = R3.valid_list_formats()
    #     for fspec in format_spec:
    #         if fspec in SO3_formats:
    #             result += self._R.to_list(fspec)
    #         elif fspec in R3_formats:
    #             result += self._x.to_list(fspec)
    #         elif fspec == "P":
    #             posemat = np.hstack((self.R().as_matrix(), self.x().as_vector()))
    #             result += posemat.ravel().tolist()
    #         elif fspec == "v":
    #             result += self._w.to_list('x')
    #         else:
    #             return NotImplemented
    #     return result
    
    @staticmethod
    def list_header(format_spec) -> list:
        result = []
        SO3_formats = SO3.valid_list_formats()
        R3_formats = R3.valid_list_formats()
        for fspec in format_spec:
            if fspec == "P":
                result += "P11,P12,P13,P14,P21,P22,P23,P24,P31,P32,P33,P34".split()
            elif fspec in R3_formats:
                result += R3.list_header(fspec)
            elif fspec in SO3_formats:
                result += SO3.list_header(fspec)
            elif fspec == 'v':
                result += "v1,v2,v3".split()
            else:
                return NotImplemented
        return result

    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (5,5):
            raise ValueError
        vecOmega = SO3.vex(mat[0:3,0:3])
        vecV = mat[0:3,3]
        vecA = mat[0:3,4]
        vec = np.concatenate((vecOmega, vecV, vecA))
        return vec

    @staticmethod
    def wedge(vec : np.ndarray) -> np.ndarray:
        if not isinstance(vec, np.ndarray):
            raise TypeError
        if not vec.size == 9:
            raise ValueError
        mat = np.zeros((5,5))
        mat[0:3,0:3] = SO3.skew(vec.ravel()[0:3])
        mat[0:3,3] = vec.ravel()[3:6]
        mat[0:3,4] = vec.ravel()[6:9]
        return mat

if __name__ == "__main__":
    P = SE23()



from abc import ABC, abstractmethod
import numpy as np

class LieGroup(ABC):
    def __str__(self):
        pass

    def __repr__(self):
        return str(self.as_matrix())

    @abstractmethod
    def __mul__(self, other):
        pass
    
    def __pow__(self, n):
        if isinstance(n, int) and n >= 0:
            result = self.identity()
            for _ in range(n):
                result = result * self
            return result
        return NotImplemented

    def __matmul__(self, other):
        if isinstance(other, LieGroup):
            return self.as_matrix() @ other.as_matrix()
        elif isinstance(other, np.ndarray):
            return self.as_matrix() @ other
        return NotImplemented
    
    @abstractmethod
    def inv(self):
        pass
    
    def __truediv__(self, other):
        return self * other.inv()
    
    @abstractmethod
    def log(self):
        pass
    
    def Adjoint(self):
        raise NotImplementedError()
    
    @staticmethod
    def adjoint(lie_alg_vector):
        return NotImplementedError()

    def as_matrix(self):
        raise NotImplementedError()
    
    @staticmethod
    def identity():
        pass

    @staticmethod
    def exp(lie_alg_vector):
        pass

    @staticmethod
    def valid_list_formats():
        return []
    
    @staticmethod
    def from_list(line, format_spec):
        pass
    
    def to_list(self, format_spec):
        raise NotImplementedError()

    @staticmethod
    def list_header(format_spec):
        pass

    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        pass

    @staticmethod
    def wedge(mat : np.ndarray) -> np.ndarray:
        pass


from src.pylie import SO3
from src.pylie import SE3
from src.pylie import SOT3
from src.pylie import SE23
from src.pylie import Quaternion
from src.pylie import SE2
from src.pylie import SL2
from src.pylie import O1_2

import numpy as np
from scipy.linalg import expm, logm
import unittest

RND_REPS = 100

class Testself(unittest.TestCase):
    Grp = SO3

    def rnd_X(self,n=1):
        if n > 1:
            return [self.Grp.exp(np.random.randn(self.Grp.DIM)) for _ in range(n)]
        else:
            return self.Grp.exp(np.random.randn(self.Grp.DIM))
    
    def rnd_v(self,n=1):
        if n > 1:
            return [np.random.randn(self.Grp.DIM) for _ in range(n)]
        else:
            return np.random.randn(self.Grp.DIM)


    def test_wedge_vee(self):
        for _ in range(RND_REPS):
            v = self.rnd_v()
            np.testing.assert_equal(self.Grp.vee(self.Grp.wedge(v)), v)

    def test_identity(self):
        for _ in range(RND_REPS):
            X = self.rnd_X()
            I = self.Grp.identity()
            np.testing.assert_almost_equal(X.as_matrix(), (X*I).as_matrix())
            np.testing.assert_almost_equal(X.as_matrix(), (I*X).as_matrix())

    def test_product(self):
        for _ in range(RND_REPS):
            X1,X2,X3 = self.rnd_X(3)
            Y1 = (X1*X2)*X3
            Y2 = X1*(X2*X3)
            np.testing.assert_almost_equal(Y1.as_matrix(), Y2.as_matrix())

    def test_matrix_product(self):
        for _ in range(RND_REPS):
            X1,X2 = self.rnd_X(2)
            Y1 = (X1*X2).as_matrix()
            Y2 = X1.as_matrix() @ X2.as_matrix()
            np.testing.assert_almost_equal(Y1, Y2)

    def test_matrix_inverse(self):
        for _ in range(RND_REPS):
            X = self.rnd_X()
            Y1 = X.inv().as_matrix()
            Y2 = np.linalg.inv(X.as_matrix())
            np.testing.assert_almost_equal(Y1, Y2)
    
    def test_matrix_identity(self):
        for _ in range(RND_REPS):
            I = self.Grp.identity().as_matrix()
            np.testing.assert_almost_equal(I, np.eye(*I.shape))

    def test_Adjoint(self):
        for _ in range(RND_REPS):
            X = self.rnd_X()
            v = self.rnd_v()
            u1 = X.Adjoint() @ v

            Y1 = X * self.Grp.exp(v) * X.inv()
            Y2 = self.Grp.exp(u1)
            np.testing.assert_almost_equal(Y1.as_matrix(), Y2.as_matrix())
    
    def test_matrix_Adjoint(self):
        for _ in range(RND_REPS):
            X = self.rnd_X()
            # Test the unit vectors
            Ad_computed = np.empty((self.Grp.DIM,self.Grp.DIM))
            for k in range(self.Grp.DIM):
                e_k = np.zeros(self.Grp.DIM)
                e_k[k] = 1.
                Ad_computed[:,k] = self.Grp.vee(X.as_matrix() @ self.Grp.wedge(e_k) @ X.inv().as_matrix())
            np.testing.assert_almost_equal(Ad_computed, X.Adjoint())

            v = self.rnd_v()
            u1 = X.Adjoint() @ v
            u2 = self.Grp.vee(X.as_matrix() @ self.Grp.wedge(v) @ X.inv().as_matrix())
            np.testing.assert_almost_equal(u1, u2)

    
    def test_inverse(self):
        for _ in range(RND_REPS):
            X = self.rnd_X()
            Y1 = X * X.inv()
            Y2 = X.inv()*X
            np.testing.assert_almost_equal(Y1.as_matrix(), self.Grp.identity().as_matrix())
            np.testing.assert_almost_equal(Y2.as_matrix(), self.Grp.identity().as_matrix())

    def test_exp(self):
        for _ in range(RND_REPS):
            v = self.rnd_v()
            X1 = self.Grp.exp(v)
            X2 = self.Grp.exp(v*0.1)**10
            np.testing.assert_almost_equal(X1.as_matrix(), X2.as_matrix())

    def test_matrix_exp(self):
        for _ in range(RND_REPS):
            v = self.rnd_v()
            X1 = self.Grp.exp(v).as_matrix()
            X2 = expm(self.Grp.wedge(v))
            np.testing.assert_almost_equal(X1, X2)

    def test_log(self):
        for k in range(self.Grp.DIM):
            e_k = np.zeros(self.Grp.DIM)
            e_k[k] = 1.
            e_k2 = self.Grp.exp(e_k).log()
            np.testing.assert_almost_equal(e_k, e_k2)

        for _ in range(RND_REPS):
            v1 = 0.01*self.rnd_v()
            v2 = self.Grp.exp(v1).log()
            np.testing.assert_almost_equal(v1, v2)

    def test_matrix_log(self):
        for _ in range(RND_REPS):
            v1 = 0.01*self.rnd_v()
            X1 = self.Grp.exp(v1).as_matrix()
            v2 = self.Grp.vee(logm(X1))

            np.testing.assert_almost_equal(v1, v2)
    
    def test_matrix_adjoint(self):
        for _ in range(RND_REPS):
            v1, v2 = self.rnd_v(2)
            u1 = self.Grp.adjoint(v1) @ v2
            u2 = self.Grp.vee(self.Grp.wedge(v1) @ self.Grp.wedge(v2) - self.Grp.wedge(v2) @ self.Grp.wedge(v1))
            np.testing.assert_almost_equal(u1, u2)
            
    
class TestSE3(Testself):
    Grp = SE3

class TestSOT3(Testself):
    Grp = SOT3

class TestSE23(Testself):
    Grp = SE23

class TestQuaternion(Testself):
    Grp = Quaternion

class TestSE2(Testself):
    Grp = SE2

class TestSL2(Testself):
    Grp = SL2

class TestO1_2(Testself):
    Grp = O1_2

if __name__ == '__main__':
    unittest.main()


from .LieGroup import LieGroup
from .SO3 import SO3 as SO3
from .R3 import R3 as R3
import numpy as np

class Quaternion(LieGroup):
    DIM = 4
    def __init__(self, q = None):
        self._quat = np.array((1,0,0,0))
        if isinstance(q, np.ndarray):
            assert q.size == 4, "Quaternions must have 4 components."
            self._quat = q.copy().reshape(4)
        elif isinstance(q, Quaternion):
            self._quat = q._quat
    
    def q(self) -> np.ndarray:
        return self._quat
    
    def real(self) -> float:
        return self._quat[0]
    
    def imag(self) -> np.ndarray:
        return self._quat[1:]
    
    def as_rotation(self):
        r = self._quat[0]
        v = self._quat[1:]
        rot = v.reshape((-1,1)) @ v.reshape((1,-1)) + r**2 * np.eye(3) + SO3.skew(v) @ SO3.skew(v)
        return SO3(rot)
    
    def as_vector(self):
        return self.q().copy()

    def __str__(self):
        return str(self._quat)
    
    def Adjoint(self):
        norm2_q = np.inner(self._quat, self._quat)
        r = self._quat[0]
        u = self._quat[1:]
        Ad = np.zeros((4,4))
        Ad[0,0] = 1.
        Ad[1:,1:] = np.eye(3) +2* (r*SO3.skew(u) + SO3.skew(u)@ SO3.skew(u)) / norm2_q
        return Ad
    
    @staticmethod
    def adjoint(quat_vec : np.ndarray) -> np.ndarray:
        assert isinstance(quat_vec, np.ndarray)
        assert quat_vec.size == 4
        ad = np.zeros((4,4))
        ad[1:,1:] = 2*SO3.skew(quat_vec[1:])
        return ad
    
    def __mul__(self, other):
        if isinstance(other, Quaternion):
            q3 = np.zeros(4)
            q3[0] = self._quat[0]*other._quat[0] - self._quat[1:] @ other._quat[1:]
            q3[1:] = self._quat[0]*other._quat[1:] + other._quat[0]*self._quat[1:] + np.cross(self._quat[1:], other._quat[1:])
            return Quaternion(q3)
        elif isinstance(other, np.ndarray) and other.size == 3:
            other_q = np.zeros(4)
            other_q[1:] = other
            result = self * other_q * self.inv()
            return result[1:]
        elif isinstance(other, np.ndarray) and other.size == 4:
            return self * Quaternion(other)
        elif isinstance(other, R3):
            result = R3()
            result._trans = self * other._trans
            return result
        return NotImplemented
    
    def __truediv__(self, other):
        if isinstance(other, Quaternion):
            return self * other.inv()
        return NotImplemented
    
    def inv(self):
        norm_sq = np.inner(self._quat, self._quat)
        result = self._quat.copy()
        result[1:] = - self._quat[1:]
        result = result / norm_sq
        return Quaternion(result)
    
    def log(self):
        r = self._quat[0]
        u = self._quat[1:]
        norm = np.linalg.norm(self._quat)
        if norm == 0:
            return np.nan * np.empty(4)
        result = np.zeros(4)
        result[0] = np.log(norm)
        norm_u = np.linalg.norm(u)
        if norm_u == 0:
            return result
        result[1:] = np.arctan(norm_u / r) * u / norm_u
        return result
    
    def as_matrix(self):
        q = self._quat
        mat = np.array((
            (q[0], q[1], q[2], q[3]),
            (-q[1], q[0], -q[3], q[2]),
            (-q[2], q[3], q[0], -q[1]),
            (-q[3],-q[2],q[1],q[0])
        ))
        return mat

    @staticmethod
    def from_matrix(mat : np.ndarray) -> 'SO3':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (4,4):
            raise ValueError
        return Quaternion(mat[0,:])
    
    @staticmethod
    def identity():
        return Quaternion(np.array((1,0,0,0)))

    @staticmethod
    def exp(quat_vec):
        assert quat_vec.size == 4, "The so(3) Lie algebra vector must have 4 elements."
        w = quat_vec[1:]
        r = quat_vec[0]
        theta = np.linalg.norm(w)
        q = np.zeros(4)
        if theta == 0:
            q[0] = 1.
        else:
            q[0] = np.cos(theta)
            q[1:] = np.sin(theta)*(w/theta)
        return Quaternion(np.exp(r)*q)

    @staticmethod
    def valid_list_formats() -> dict:
        # Possible formats are
        # q : 4 entry quaternion (scalar last)
        # w : 4 entry quaternion (scalar first)
        return {'q':4, 'w':4}

    @staticmethod
    def from_list(line, format_spec="q") -> 'Quaternion':
        result = Quaternion()
        if format_spec == "q":
            result._quat = np.array([float(line[i]) for i in range(4)])
            line = line[4:]
        elif format_spec == "w":
            result._quat = np.array([float(line[i]) for i in [1,2,3,0]])
            line = line[4:]
        else:
            return NotImplemented
        return result

    def to_list(self, format_spec) -> list:
        if format_spec == "q":
            temp = self._quat.tolist()
            result = [temp[i] for i in [1,2,3,0]]
        elif format_spec == "w":
            result = self._quat.tolist()
        else:
            return NotImplemented
        return result
    
    @staticmethod
    def list_header(format_spec):
        if format_spec == "q":
            result = "qx,qy,qz,qw".split()
        elif format_spec == "w":
            result = "qw,qx,qy,qz".split()
        else:
            return NotImplemented
        return result

    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (4,4):
            raise ValueError
        return mat[0,:]

    @staticmethod
    def wedge(vec : np.ndarray) -> np.ndarray:
        if not isinstance(vec, np.ndarray):
            raise TypeError
        if not vec.size == 4:
            raise ValueError
        return Quaternion(vec).as_matrix()
    
    


try:
    import matplotlib.pyplot as plt
    from matplotlib.artist import Artist
    from matplotlib.patches import Wedge, Circle, Arc
    from matplotlib.axes import Axes
    from mpl_toolkits.mplot3d import Axes3D
except ImportError as error:
    print("The plotting functionality of pylie depends on matplotlib.")
    raise error

from pylie import SO3, SE3, SIM3, SE2, LieGroup
import numpy as np

class frameArtist(Artist):
    def __init__(self, ax : Axes3D, pose_data : SE3, style='-', colors=None, length=1.0):
        if not isinstance(ax, Axes3D):
            raise TypeError("The axes must be of 3D type.")
        if not isinstance(pose_data, (SIM3, SE3, SO3)):
            raise TypeError("The pose_data must be of one of the types: SIM3, SE3, SO3.")

        self._pose_data = pose_data
        self._length = length
        Q, t = self._pose_to_Qt(pose_data)

        if colors is None:
            colors = ['r','g','b','k']
        elif isinstance(colors,str):
            colors = [colors]*4
        self._frame_lines = []
        for a in range(3):
            temp, = ax.plot([t[0,0], t[0,0]+length*Q[0,a]], [t[1,0], t[1,0]+length*Q[1,a]], [t[2,0], t[2,0]+length*Q[2,a]], linestyle=style, color=colors[a])
            self._frame_lines.append(temp)
        self._frame_center, = ax.plot(t[0,0], t[1,0], t[2,0], colors[3]+'.')

        if isinstance(pose_data, SO3):
            ax.set_xlim(-1, 1)
            ax.set_ylim(-1, 1)
            ax.set_zlim(-1, 1)
    
    def set_pose_data(self, pose : SE3):
        if not isinstance(pose, (SIM3, SE3, SO3)):
            raise TypeError("The frame must be of one of the types: SIM3, SE3, SO3.")
        
        self._pose_data = pose
        Q, t = self._pose_to_Qt(pose)
        for a in range(3):
            self._frame_lines[a].set_data_3d([t[0,0], t[0,0]+self._length*Q[0,a]], [t[1,0], t[1,0]+self._length*Q[1,a]], [t[2,0], t[2,0]+self._length*Q[2,a]])
        self._frame_center.set_data_3d([t[0,0]], [t[1,0]], [t[2,0]])

    
    @staticmethod
    def _pose_to_Qt(pose : SE3):
        if isinstance(pose, SO3):
            mat = np.eye(4)
            mat[0:3,0:3] = pose.as_matrix()
        else:
            mat = pose.as_matrix()
        
        Q = mat[0:3,0:3]
        t = mat[0:3,3:4]
        return Q, t
    

class frameArtist2(Artist):
    # Frame artist for SE(2) frames
    def __init__(self, ax : Axes, pose_data : SE2, style='-', colors=None, length=1.0):
        if not isinstance(ax, Axes):
            raise TypeError("The axes must be of 2D type.")
        if not isinstance(pose_data, (SE2)):
            raise TypeError("The pose_data must be of SE2 type.")

        self._pose_data = pose_data
        self._length = length
        Q, t = pose_data.R(), pose_data.p()

        if colors is None:
            colors = ['r','g','b']
        elif isinstance(colors,str):
            colors = [colors]*3
        self._frame_lines = []
        for a in range(2):
            temp, = ax.plot([t[0], t[0]+length*Q[0,a]], [t[1], t[1]+length*Q[1,a]], linestyle=style, color=colors[a])
            self._frame_lines.append(temp)
        self._frame_center, = ax.plot(t[0], t[1], colors[-1]+'.')
    
    def set_pose_data(self, pose : SE2):
        if not isinstance(pose, (SE2)):
            raise TypeError("The pose_data must be of SE2 type.")
        
        self._pose_data = pose
        Q, t = pose.R(), pose.p()
        for a in range(2):
            self._frame_lines[a].set_data([t[0], t[0]+self._length*Q[0,a]], [t[1], t[1]+self._length*Q[1,a]])
        self._frame_center.set_data([t[0]], [t[1]])

    
    @staticmethod
    def _pose_to_Qt(pose : SE2):
        return pose.theta(), pose.p()

class ArtificialHorizonArtist(Artist):
    def __init__(self, ax : plt.Axes, attitude_data : SO3):
        if not isinstance(ax, plt.Axes):
            raise TypeError("The axes must be of plt.Axes type.")
        if not isinstance(attitude_data, SO3):
            raise TypeError("The attitude_data must be an SO3 element.")

        # Background
        ah_circle = Circle((0,0), radius=1.0, color='w', ec='w')
        ah_colouring = Circle((0,0), radius=5.0, color='k')
        ax.add_patch(ah_colouring)
        ax.add_patch(ah_circle)


        # Horizon line
        self._attitude_data = attitude_data
        true_horizon = self._compute_horizon(attitude_data)
        self._horizon_line, = ax.plot(true_horizon[0,:], true_horizon[1,:])

        # Horizon shading
        height = self._compute_horizon_height(attitude_data)
        slope = self._compute_horizon_slope(attitude_data) * 180.0 / np.pi
        self._shade = Wedge((0,height), 2.0, -180+slope, slope, alpha=0.5)
        ax.add_patch(self._shade)

        # Clip path
        clip_circle = Circle((0, 0), radius=1.0, transform=ax.transData)
        self._horizon_line.set_clip_path(clip_circle)
        self._shade.set_clip_path(clip_circle)

        # Center Overlay
        ax.add_patch(Arc((0,0), 0.3, 0.3, theta1=180, lw=2.0))
        ax.plot([0.15, 0.6], [0,0], color='k', lw=2.0)
        ax.plot([-0.15, -0.6], [0,0], color='k', lw=2.0)

        # Axes settings
        ax.axis('square')
        ax.set_xlim([-1,1])
        ax.set_ylim([-1,1])
    
    def set_attitude_data(self, attitude : SE3):
        if not isinstance(attitude, SO3):
            raise TypeError("The attitude_data must be an SO3 element.")
        
        self._attitude_data = attitude
        true_horizon = self._compute_horizon(attitude)
        self._horizon_line.set_data(true_horizon[0,:], true_horizon[1,:])

        height = self._compute_horizon_height(attitude)
        slope = self._compute_horizon_slope(attitude) * 180.0 / np.pi
        self._shade.set_center((0,height))
        self._shade.set_theta1(-180 + slope)
        self._shade.set_theta2(slope)
    
    @staticmethod
    def _compute_horizon(attitude : SO3):
        angles = np.linspace(0, 2*np.pi, 50)
        base_horizon = np.vstack((np.cos(angles), np.sin(angles), np.zeros(angles.shape)))
        true_horizon = attitude.inv() * base_horizon
        true_horizon = true_horizon[:, true_horizon[0,:]>0]
        true_horizon = true_horizon / true_horizon[0,:]
        true_horizon = true_horizon[1:,:]
        return true_horizon
    
    @staticmethod
    def _compute_horizon_slope(attitude : SO3):
        theta = ArtificialHorizonArtist._find_heading_angle(attitude)
        H = attitude.inv() * np.array([[np.cos(theta)], [np.sin(theta)], [0.0]])
        DH = attitude.inv() * np.array([[-np.sin(theta)], [np.cos(theta)], [0.0]])
        DH2 = H[0,0] * DH - DH[0,0] * H
        slope = np.arctan2(DH2[2,0], DH2[1,0])
        return slope
    
    @staticmethod
    def _compute_horizon_height(attitude : SO3):
        theta = ArtificialHorizonArtist._find_heading_angle(attitude)
        H = attitude.inv() * np.array([[np.cos(theta)], [np.sin(theta)], [0.0]])
        H = H / H[0,0]
        return H[2,0]
    
    @staticmethod
    def _find_heading_angle(attitude):
        RT = attitude.as_matrix().T
        theta = np.arctan2(-RT[1,0], RT[1,1])
        return theta




def plotFrame(frame : LieGroup, style='-', colors=None, ax : Axes3D = None, length=1.0):
    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot(111, projection='3d')
    
    frame_artist = frameArtist(ax, frame, style, colors, length)

    return frame_artist

def plotFrame2(frame : SE2, style='-', colors=None, ax : Axes = None, length=1.0):
    if ax is None:
        fig = plt.gcf()
        ax = fig.add_subplot()
    
    frame_artist = frameArtist2(ax, frame, style, colors, length)

    return frame_artist

def plotArtificialHorizon(attitude : SO3, ax : Axes3D = None):
    if ax is None:
        ax = plt.gca()
    horizon_artist = ArtificialHorizonArtist(ax, attitude)
    return horizon_artist

if __name__ == "__main__":
    # pose = SE3.identity()
    pose = SE3.from_list([1,2,3], "x")
    pose2 = SE3.from_list([1.5,2.5,3.5], "x")

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    artist = plotFrame(pose)
    plt.pause(1)

    artist.set_pose_data(pose2)
    plt.show()

    plt.clf()

    yaw_tf = SO3.from_list(np.array([0,0,0.5]), 'r')
    att = SO3.from_list(np.array([0.1,0,0.]), 'r')
    att2 = SO3.from_list(np.array([0.2,0,0]), 'r')
    for i in range(10):
        plt.cla()
        att = SO3.from_list(np.array([0.1*i,0.0,0]), 'r')
        artist = plotArtificialHorizon(att)
        plt.pause(0.2)

    for i in range(10):
        plt.cla()
        att = SO3.from_list(np.array([0.0,-0.05*i,0]), 'r')
        artist = plotArtificialHorizon(att)
        plt.pause(0.2)
    
    plt.show()

    plt.clf()


    pose = SE2(0.2, np.array([0.1,0.5]))
    pose2 = SE2(0.3, np.array([0.1,0.5]))
    artist = plotFrame2(pose)
    plt.pause(1)
    artist.set_pose_data(pose2)
    plt.show()





from .LieGroup import LieGroup
import numpy as np

class R3(LieGroup):
    # The Lie group of dim 3 translation.
    # [ I_3  x ]
    # [  0   1 ]
    def __init__(self, x = None):
        if x is None:
            self._trans = np.zeros(3)
        elif isinstance(x, R3):
            self._trans = x._trans
        elif isinstance(x, np.ndarray) or isinstance(x ,list):
            self._trans = np.reshape(x, 3)
        else:
            self._trans = np.zeros(3)
    
    def x(self) -> np.ndarray:
        return self._trans
    
    def __str__(self):
        return str(self._trans.ravel())
    
    def Adjoint(self):
        return np.eye(3)
    
    def __mul__(self, other):
        if isinstance(other, R3):
            result = R3()
            result._trans = self._trans + other._trans
            return result
        elif isinstance(other, np.ndarray):
            if other.shape[0] == 3:
                return other + self._trans
            elif other.shape[0] == 4:
                return self.as_matrix() @ other
        elif isinstance(other, float):
            return R3(self._trans * other)
        return NotImplemented
    
    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __add__(self, other):
        return self * other
    
    def __neg__(self):
        return self.inv()
    
    def __truediv__(self, other):
        if isinstance(other, R3):
            result = R3()
            result._trans = self._trans - other._trans
            return result
        return NotImplemented
    
    def __sub__(self, other):
        return self / other

    def inv(self):
        result = R3()
        result._trans = -self._trans
        return result
    
    def log(self):
        return self._trans
    
    def as_matrix(self):
        result = np.eye(4)
        result[0:3,3:4] = self._trans
        return result
    
    def as_vector(self):
        return self._trans.copy()
    
    @staticmethod
    def identity():
        result = R3()
        result._trans = np.zeros(3)
        return result

    @staticmethod
    def exp(tr3vec):
        assert tr3vec.size == (3), "Invalid shape of Lie algebra vector."
        result = R3()
        result._trans = tr3vec
        return result

    @staticmethod
    def valid_list_formats() -> dict:
        # Possible formats are
        # x : 3 entry vector
        return {'x':3}

    @staticmethod
    def from_list(line, format_spec="x") -> 'R3':
        result = R3()
        if format_spec == "x":
            result._trans = np.array([float(line[i]) for i in range(3)])
            line = line[3:]
        else:
            return NotImplemented
        return result

    def to_list(self, format_spec) -> list:
        if format_spec == "x":
            result = self._trans.ravel().tolist()
        else:
            return NotImplemented
        return result
    
    @staticmethod
    def list_header(format_spec):
        if format_spec == "x":
            result = "x1,x2,x3".split()
        else:
            return NotImplemented
        return result

if __name__ == "__main__":
    x = R3()


from .LieGroup import LieGroup
from .SO3 import SO3 as SO3
from .MR1 import MR1 as MR1
import numpy as np

class SOT3(LieGroup):
    DIM = 4
    def __init__(self, R : SO3 = None, s : MR1 = None):
        if R is None:
            R = SO3()
        if s is None:
            s = MR1()
        self._R = SO3(R)
        self._s = MR1(s)
    
    def R(self) -> np.ndarray:
        return self._R
    
    def s(self) -> np.ndarray:
        return self._s

    def __str__(self):
        return str(self.as_matrix())
    
    def Adjoint(self):
        ad = np.eye(4)
        ad[0:3,0:3] = self._R.as_matrix()
        return ad
    
    @staticmethod
    def adjoint(sot3vec : np.ndarray) -> np.ndarray:
        assert isinstance(sot3vec, np.ndarray)
        assert sot3vec.size == 4
        ad = np.zeros((4,4))
        ad[:3,:3] = SO3.skew(sot3vec[:3])
        return ad

    def __mul__(self, other):
        if isinstance(other, SOT3):
            result = SOT3()
            result._R = self._R * other._R
            result._s = self._s * other._s
            return result
        if isinstance(other, np.ndarray):
            if other.shape[0] == 3:
                return self._s * (self._R * other)
            elif other.shape[0] == 4:
                return self.as_matrix() @ other
        
        return NotImplemented
    
    @staticmethod
    def identity():
        result = SOT3()
        result._R = SO3.identity()
        result._s = MR1.identity()
        return result
    
    def as_matrix(self):
        mat = self._s.as_float() * self._R.as_matrix()
        return mat
    
    def as_quaternion(self) -> np.ndarray:
        return self.R().as_quaternion() * self._s

    @staticmethod
    def from_matrix(mat : np.ndarray) -> 'SOT3':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not (mat.shape == (4,4) or mat.shape == (3,3)):
            raise ValueError
        
        result = SOT3()
        m = mat[0:3,0:3]
        Q = m.T @ m
        # Q is s^2 I_3.
        result._s = MR1(Q[0,0]**0.5)
        result._R = SO3.from_matrix(m / result._s)

        return result
    
    def __truediv__(self, other):
        if isinstance(other, SOT3):
            return self * other.inv()
        return NotImplemented
    
    def inv(self):
        result = SOT3()
        result._R = self._R.inv()
        result._s = self._s.inv()
        return result
    
    @staticmethod
    def exp(sot3arr) -> 'SOT3':
        if not isinstance(sot3arr, np.ndarray):
            raise TypeError
        if sot3arr.shape == (4,4):
            sot3arr = SOT3.vee(sot3arr)
        elif not sot3arr.size == 4:
            raise ValueError

        result = SOT3(SO3.exp(sot3arr.ravel()[0:3]), np.exp(sot3arr.item(3)))

        return result
    
    def log(self) -> np.ndarray:
        w = self._R.log()
        a = self._s.log()
        return np.concatenate((w,(a,)))

    @staticmethod
    def valid_list_formats() -> dict:
        # Possible formats are
        # q/w/R/r : SO(3) format specs
        # s : 1 entry scale
        # Q : 9 entry matrix (row-by-row)
        result = {'Q':9}
        result.update(SO3.valid_list_formats())
        result.update(MR1.valid_list_formats())
        return result

    @staticmethod
    def from_list(line, format_spec="qs") -> 'SOT3':
        result = SOT3()
        SO3_formats = SO3.valid_list_formats()
        MR1_formats = MR1.valid_list_formats()
        for fspec in format_spec:
            if fspec in SO3_formats:
                result._R = SO3.from_list(line, fspec)
                line = line[SO3_formats[fspec]:]
            elif fspec in MR1_formats:
                result._s = MR1.from_list(line, fspec)
                line = line[MR1_formats[fspec]:]
            elif fspec == "Q":
                mat = np.reshape(np.array([float(line[i]) for i in range(9)]), (3,3))
                result = SOT3.from_matrix(mat)
                line = line[9:]
            else:
                return NotImplemented
        return result

    def to_list(self, format_spec) -> list:
        result = []
        SO3_formats = SO3.valid_list_formats()
        MR1_formats = MR1.valid_list_formats()
        for fspec in format_spec:
            if fspec in SO3_formats:
                result += self._R.to_list(fspec)
            elif fspec in MR1_formats:
                result += self._s.to_list(fspec)
            elif fspec == "Q":
                result += self.as_matrix().ravel().tolist()
            else:
                return NotImplemented
        return result
    
    @staticmethod
    def list_header(format_spec) -> list:
        result = []
        SO3_formats = SO3.valid_list_formats()
        MR1_formats = MR1.valid_list_formats()
        for fspec in format_spec:
            if fspec == "Q":
                result += "Q11,Q12,Q13,Q21,Q22,Q23,Q31,Q32,Q33".split()
            elif fspec in MR1_formats:
                result += MR1.list_header(fspec)
            elif fspec in SO3_formats:
                result += SO3.list_header(fspec)
            else:
                return NotImplemented
        return result

    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (3,3):
            raise ValueError
        vec = np.zeros(4)
        vec[0:3] = SO3.vex(mat)
        vec[3] = mat[0,0]
        return vec

    @staticmethod
    def wedge(vec : np.ndarray) -> np.ndarray:
        if not isinstance(vec, np.ndarray):
            raise TypeError
        if not vec.size == 4:
            raise ValueError
        result = SO3.skew(vec.ravel()[0:3]) + np.identity(3) * vec.item(3)
        return result

if __name__ == "__main__":
    P = SOT3()


from .LieGroup import LieGroup
import numpy as np

class SL2(LieGroup):
    DIM = 3
    def __init__(self, H : np.ndarray = None):
        if H is None:
            self._H = np.eye(2)
        else:
            assert H.shape == (2,2), "H must be a 2x2 matrix"
            self._H = H
    
    def H(self) -> np.ndarray:
        return self._H

    def __str__(self) -> str:
        return str(self.as_matrix())
    
    def Adjoint(self) -> np.ndarray:
        a = self._H[0,0]
        b = self._H[0,1]
        c = self._H[1,0]
        d = self._H[1,1]
        return np.array([
            [2*b*c+1, -a*c, b*d],
            [-2*a*b, a*a, -b*b],
            [2*c*d, -c*c, d*d]
        ])
    
    @staticmethod
    def adjoint(sl2vec : np.ndarray) -> np.ndarray:
        assert isinstance(sl2vec, np.ndarray)
        assert sl2vec.size == 3
        u1 = sl2vec[0]
        u2 = sl2vec[1]
        u3 = sl2vec[2]
        return np.array([
            [0, -u3, u2],
            [-2*u2, 2*u1, 0],
            [2*u3, 0, -2*u1]
        ])
    
    def __mul__(self, other) -> 'SL2':
        if isinstance(other, SL2):
            result = SL2(self._H @ other._H)
            return result
        if isinstance(other, np.ndarray):
            if other.shape[0] == 2:
                return self.H @ other
        
        return NotImplemented
    
    @staticmethod
    def identity() -> 'SL2':
        return SL2(np.eye(2))
    
    def as_matrix(self) -> np.ndarray:
        return self._H

    @staticmethod
    def from_matrix(mat : np.ndarray) -> 'SL2':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (2,2):
            raise ValueError
        
        result = SL2(mat)
        return result
    
    def __truediv__(self, other) -> 'SL2':
        if isinstance(other, SL2):
            return self * other.inv()
        return NotImplemented
    
    def inv(self) -> 'SL2':
        return SL2(np.array((
                (self._H[1,1], -self._H[0,1]),
                (-self._H[1,0], self._H[0,0])
            )))
    
    @staticmethod
    def exp(sl2arr) -> 'SL2':
        if not isinstance(sl2arr, np.ndarray):
            raise TypeError
        if sl2arr.shape == (2,2):
            U = SL2.vee(sl2arr)
        elif sl2arr.size == 3:
            U = sl2arr
        else:
            raise ValueError
        
        a = U.item(0)
        b = U.item(1)
        c = U.item(2)

        theta = a**2 + b*c
        if theta > 1e-8:
            theta_sqrt = np.sqrt(theta)
            A = np.cosh(theta_sqrt)
            B = np.sinh(theta_sqrt)/theta_sqrt
        elif theta < -1e-8:
            theta_sqrt = np.sqrt(-theta)
            A = np.cos(theta_sqrt)
            B = np.sin(theta_sqrt)/theta_sqrt
        else:
            A = 1.0
            B = 1.0
        H = A * np.eye(2) + B * SL2.wedge(U)

        return SL2(H)
    
    def log(self) -> np.ndarray:
        H = self._H
        alpha = 0.5*np.trace(H)
        if alpha > 1+1e-8:
            tmp = np.arccosh(alpha)
            A = tmp / np.sinh(tmp)
        elif alpha < 1-1e-8:
            tmp = np.arccos(alpha)
            A = tmp / np.sin(tmp)
        else:
            A = 1.

        U_wedge = A * (H - alpha*np.eye(2))
        return SL2.vee(U_wedge)
        
    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (2,2):
            raise ValueError

        return np.array((mat[0,0], mat[0,1], mat[1,0]))

    @staticmethod
    def wedge(vec : np.ndarray) -> np.ndarray:
        if not isinstance(vec, np.ndarray):
            raise TypeError
        if not vec.size == 3:
            raise ValueError
        mat = np.array((
            (vec.item(0), vec.item(1)),
            (vec.item(2), -vec.item(0))
        ))
        return mat

if __name__ == "__main__":
    P = SL2()



from .LieGroup import LieGroup
from .R3 import R3 as R3
from scipy.spatial.transform import Rotation
import numpy as np

class SO3(LieGroup):
    DIM = 3
    def __init__(self, R = None):
        if R is None:
            self._R = np.eye(3)
        elif isinstance(R, np.ndarray):
            self._R = R
        elif isinstance(R, SO3):
            self._R = R._R
        elif isinstance(R, Rotation):
            self._R = R.as_matrix()
            
    @staticmethod
    def adjoint(so3vec : np.ndarray) -> np.ndarray:
        assert isinstance(so3vec, np.ndarray)
        assert so3vec.size == 3
        return SO3.skew(so3vec)
            
    def R(self) -> np.ndarray:
        return self._R

    def as_quaternion(self) -> np.ndarray:
        return Rotation.from_matrix(self._R).as_quat()

    def as_euler(self, seq='xyz', degrees=True) -> np.ndarray:
        return Rotation.from_matrix(self._R).as_euler(seq, degrees=degrees)

    @staticmethod
    def from_euler(euler_angles, seq='xyz', degrees=True) -> np.ndarray:
        return SO3.from_matrix(Rotation.from_euler(seq=seq, angles=euler_angles, degrees=degrees).as_matrix())

    def __str__(self):
        return str(self._R)
    
    def Adjoint(self):
        return self.as_matrix()
    
    def __mul__(self, other):
        if isinstance(other, SO3):
            return SO3(self._R @ other._R)
        elif isinstance(other, np.ndarray) and other.shape[0] == 3:
            return self.as_matrix() @ other
        elif isinstance(other, R3):
            result = R3()
            result._trans = self._R @ other._trans
            return result
        return NotImplemented
    
    def __truediv__(self, other):
        if isinstance(other, SO3):
            return self * other.inv()
        return NotImplemented
    
    def inv(self):
        return SO3(self._R.T)
    
    def log(self):
        theta = np.arccos((np.trace(self._R) - 1.0) / 2.0)
        if abs(theta) > 1e-6:
            coefficient = theta / (2.0 * np.sin(theta))
        else:
            coefficient = 0.5

        Omega = coefficient * (self._R - self._R.transpose())
        return SO3.vee(Omega)
    
    def as_matrix(self):
        return self._R

    @staticmethod
    def from_matrix(mat : np.ndarray) -> 'SO3':
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (3,3):
            raise ValueError
        
        return SO3(mat)
    
    @staticmethod
    def identity():
        return SO3(np.eye(3))

    @staticmethod
    def exp(so3vec):
        assert so3vec.size == 3, "The so(3) Lie algebra vector must have 3 elements."
        theta = np.linalg.norm(so3vec)
        if theta == 0.0:
            return SO3(np.eye(3))
        elif theta < 1e-6:
            t1 = 1.0 - theta**2/6.0
            t2 = 0.5 - theta**2/24.0
        else:
            t1 = np.sin(theta) / theta
            t2 = (1.0 - np.cos(theta)) / (theta**2)

        Omega = SO3.wedge(so3vec)
        return SO3(np.eye(3) + t1 * Omega + t2 * Omega @ Omega)

    @staticmethod
    def valid_list_formats() -> dict:
        # Possible formats are
        # R : 9 entry matrix (row-by-row)
        # q : 4 entry quaternion (scalar last)
        # w : 4 entry quaternion (scalar first)
        # r : 3 entry log vector
        return {'R':9, 'q':4, 'w':4, 'r':3}

    @staticmethod
    def from_list(line, format_spec="q") -> 'SO3':
        result = SO3()
        if format_spec == "R":
            mat = np.reshape(np.array([float(line[i]) for i in range(9)]), (3,3))
            result._R = mat
            line = line[9:]
        elif format_spec == "q":
            quat = np.array([float(line[i]) for i in range(4)])
            result._R = Rotation.from_quat(quat).as_matrix()
            line = line[4:]
        elif format_spec == "w":
            quat = np.array([float(line[i]) for i in [1,2,3,0]])
            result._R = Rotation.from_quat(quat).as_matrix()
            line = line[4:]
        elif format_spec == "r":
            rotvec = np.array([float(line[i]) for i in range(3)])
            result = SO3.exp(rotvec)
            line = line[3:]
        else:
            return NotImplemented
        return result

    def to_list(self, format_spec) -> list:
        if format_spec == "R":
            mat = self.as_matrix()
            result = mat.ravel().tolist()
        elif format_spec == "q":
            quat = self.as_quat()
            result = quat.ravel().tolist()
        elif format_spec == "w":
            quat = self.as_quat()
            temp = quat.ravel().tolist()
            result = [temp[i] for i in [3,0,1,2]]
        elif format_spec == "r":
            rotvec = self.log()
            result = rotvec.ravel().tolist()
        else:
            return NotImplemented
        return result
    
    @staticmethod
    def list_header(format_spec):
        if format_spec == "R":
            result = "R11,R12,R13,R21,R22,R23,R31,R32,R33".split()
        elif format_spec == "q":
            result = "qx,qy,qz,qw".split()
        elif format_spec == "w":
            result = "qw,qx,qy,qz".split()
        elif format_spec == "r":
            result = "rx,ry,rz".split()
        else:
            return NotImplemented
        return result

    @staticmethod
    def vee(mat : np.ndarray) -> np.ndarray:
        if not isinstance(mat, np.ndarray):
            raise TypeError
        if not mat.shape == (3,3):
            raise ValueError
        vec = np.array([mat[2,1],
                        mat[0,2],
                        mat[1,0]])
        return vec
    
    @staticmethod
    def vex(mat : np.ndarray) -> np.ndarray:
        return SO3.vee(mat)

    @staticmethod
    def wedge(vec : np.ndarray) -> np.ndarray:
        if not isinstance(vec, np.ndarray):
            raise TypeError
        if not vec.size == 3:
            raise ValueError
        mat = np.array([[      0.0, -vec.item(2),  vec.item(1)],
                        [ vec.item(2),       0.0, -vec.item(0)],
                        [-vec.item(1),  vec.item(0),       0.0]])
        return mat
    
    @staticmethod
    def skew(vec : np.ndarray) -> np.ndarray:
        return SO3.wedge(vec)
    
    @staticmethod
    def from_vectors(origin : np.ndarray, dest : np.ndarray) -> 'SO3':
        origin = origin / np.linalg.norm(origin)
        dest = dest / np.linalg.norm(dest)
        v = np.cross(origin, dest, axis=0);
        c = origin.T @ dest
        s = np.linalg.norm(v)

        if (abs(1 + c) <= 1e-8):
            # The vectors are exactly opposing
            # We need to find a vector perpendicular to origin
            # e1 will work almost always. Otherwise e2 will.
            v2 = np.reshape((1.0,0,0), (3,1))
            if abs(float(v2.T @ origin)) > 1 - 1e-8:
                v2 = np.reshape((0,1.0,0), (3,1))
            v2 = np.cross(origin, v2, axis=0) # This is perpendicular to origin and not zero
            v2 = v2 / np.linalg.norm(v2)
            # Rotating around v2 by pi rads will give the desired result
            R = SO3.exp(np.pi * v2)

        else:
            vx = SO3.skew(v)
            mat = np.eye(3) + vx + (1 - c) / (s**2) * vx @ vx;
            R = SO3.from_matrix(mat)

        return R

if __name__ == "__main__":
    R = SO3()